"""Module containing universal functions for file operations that can be used with urpa robots"""

import glob
import logging
import os
import shutil
import time
import traceback

from typing import Optional

from .universal import timestamp

import __main__

logger = logging.getLogger(__name__)


def remove_dir(path: str) -> None:
    """Removes a directory

    :param path:    path
    :return:        None
    """
    if os.path.isdir(path):
        logger.info(f"Removing directory '{path}'")
        shutil.rmtree(path)


def remove(path: str) -> None:
    """Removes a file

    :param path:    path
    :return:        None
    """
    if os.path.isfile(path):
        os.remove(path)


def move(file_path: str, destination: str) -> None:
    """Moves a file

    :param file_path:    path
    :param destination:  destination path
    :return:             None
    """
    if os.path.isfile(file_path):
        shutil.move(file_path, destination)


def copy(src: str, dest: str) -> None:
    """Copies a file

    :param src:   source path
    :param dest:  destination path
    :return: None
    """
    shutil.copyfile(src, dest)


def remove_files_older_than(dir_path: str, days: int) -> None:
    """Removes all files in a directory that are older than 'days' days

    :param dir_path:     path
    :param days:         number of days
    :return:             None
    """
    files = os.listdir(dir_path)
    for file in files:
        file_path = os.path.join(dir_path, file)
        if os.path.isfile(file_path):
            # file age in seconds
            file_age = time.time() - os.stat(file_path).st_mtime
            # file age in days
            file_age /= 86400
            if file_age > days:
                logger.info(f"Removing '{file_path}' because it is older than '{days}' days")
                remove(file_path)


def write_txt_file(file_name: str, content: str, mode: str = "w", encoding: str = "utf-8") -> None:
    """Writes a text file

    :param file_name:    path
    :param content:      string to write
    :param mode:         mode of writing (writing, appending, ...)
    :param encoding:     encoding to use
    :return:             None
    """
    possible_modes = ("w", "a", "w+", "a+")
    if not mode in possible_modes:
        raise ValueError(f"Invalid write mode '{mode}'. Please use one of the following: '{possible_modes}'")
    with open(file_name, mode, encoding=encoding) as txt_file:
        txt_file.write(content)


def read_txt_file(file_name: str, encoding: str = "utf-8") -> str:
    """Reads a text file

    :param file_name:    path
    :param encoding:     encoding to use
    :return:             str content
    """
    with open(file_name, "r", encoding=encoding) as txt_file:
        return txt_file.read()


class Helper:
    """Class for reading and writing helper text files"""

    def __init__(self, file_name: str, init_value: int = 0) -> None:
        """Init. Creates file if it does not exist

        :param file_name:      file name
        """
        self.file_name = file_name
        if not os.path.isfile(file_name):
            logger.info(f"Creating '{file_name}' file")
            self.write(init_value)

    def get(self) -> int:
        """Reads the file and returns its content as integer

        :return:    int
        """
        with open(self.file_name) as helper_file:
            return int(helper_file.read())

    def write(self, value: int) -> None:
        """Writes integer 'value' to the file

        :param value:    integer
        :return:         None
        """
        if not isinstance(value, int):
            logger.error(f"Value '{value}' is not an integer")
            raise ValueError(f"Value '{value}' is not an integer")
        with open(self.file_name, "w") as helper_file:
            helper_file.write(str(value))

    def increment(self, increment: int = 1) -> int:
        """Increments number in file by 'increment'

        :param increment:    how much to increment by
        :return:             number after incrementing
        """
        if not isinstance(increment, int):
            logger.error(f"Value '{increment}' is not an integer")
            raise ValueError(f"Value '{increment}' is not an integer")
        new_value = self.get() + increment
        self.write(new_value)
        return new_value

    def delete(self) -> None:
        """Removes the file

        :return:       None
        """
        os.remove(self.file_name)


def archivate_file(
    source_file: str,
    destination_path: str,
    prefix_timestamp_format: Optional[str] = None,
    force_rewrite: bool = False,
) -> str:
    """Moves 'source_file' to 'destination_path' and if selected adds timestamp prefix to its name

    :param source_file:              path to file
    :param destination_path:         path to directory
    :param prefix_timestamp_format:  format of the timestamp prefix. If None no prefix is added
    :param force_rewrite:            bool - if destination file already exists it is rewriten if True
    :return:                         string - path to the new file
    """
    if not os.path.isfile(source_file):
        raise FileNotFoundError(f"File '{source_file}' does not exist")

    file_name = os.path.basename(source_file)
    prefix = timestamp(prefix_timestamp_format) if prefix_timestamp_format else ""
    file_name = f"{prefix}{file_name}"
    output_file_path = os.path.join(destination_path, file_name)

    if os.path.isfile(output_file_path):
        if force_rewrite:
            logger.info(f"Rewriting file '{output_file_path}'")
        else:
            raise FileExistsError(
                f"Cannot write file '{output_file_path}' because it already exists. If you want to rewrite it use 'force_rewrite = True'"
            )

    move(source_file, output_file_path)
    return output_file_path


def prepare_dir(dir_name: str) -> None:
    """Creates directory dir_name if it does not exist

    :param dir_name:      path
    :return:              None
    """
    if not os.path.exists(dir_name):
        logger.info(f"Creating new directory '{dir_name}'.")
        os.mkdir(dir_name)


# TODO remove this.  _get_main_file_name() with the stack trace seems to work fine
# def _get_main_file_name() -> str:
#     """Parses all files in current working directory and returns name of file that contains function 'main'
#     approach via __main__.__file__ does not work because urpa robot loads files dynamicaly so file __main__
#     has no attribude __file__

#     # TODO ignore lines that are commented out
#     # TODO 2: write tests for this

#     :return:       basename of the main file
#     """
#     main_file = None
#     all_files = os.listdir()
#     py_files = [file for file in all_files if ".py" in file]
#     for file_name in py_files:
#         try:
#             with open(file_name, "r", encoding="utf-8") as file:
#                 content = file.read()
#         except PermissionError:
#             continue
#         if "def main()" in content:
#             if not main_file:
#                 main_file = os.path.splitext(file_name)[0]
#             else:
#                 raise RuntimeError(
#                     "More than one .py file contains function 'main()'. Please specify 'current_log_dir' arg manually"
#                 )
#     if not main_file:
#         raise RuntimeError(
#             "No .py file containing function 'main()' found. Please specify 'current_log_dir' arg manually"
#         )
#     return main_file


def _get_main_file_name() -> str:
    """Gets name of the main file from stack trace

    :return:    basename of the main file without an extension
    """
    for line in traceback.format_stack():
        if "File" in line and ".py" in line:
            # first line that contains 'File' and '.py' is the main file
            # e.g. 'File "c:\path\to\file.py", line 12, in <module>'
            break
    # parse only the path part from the string
    indexes_of_file_path = [i for i, x in enumerate(line) if x == '"']
    main_file_path = line[indexes_of_file_path[0] + 1 : indexes_of_file_path[1]]
    main_file_name = os.path.basename(main_file_path)
    return os.path.splitext(main_file_name)[0]


def copy_error_img(
    output_dir: str,
    output_file_name: Optional[str] = None,
    screenshot_format: str = "png",
    current_log_dir: str = "",
    offset: int = 0,
) -> str:
    r"""Finds 'screenshot_format' file in the 'current_log_dir' and copies it to 'output_dir'.
    'offset' is used to determine which file to copy starting from the last
        offset=0 -> last file, offset=1 -> second last file, ...
    Files are ordered by their age in descending order. Last file (offset 0) is always the newest one

    :param output_dir:           path
    :param output_file_name:     optional name of the copied file. Name of the original is used if none provided
    :param screenshot_format:    png, or bmp
    :param current_log_dir:      directory containing the screenshots. Defaults to 'log\main_module_name_YYYY-MM-DD'
    :param offset:               which file to copy, starting from last one
    :return:                     str path to copied file
    """
    if not current_log_dir:
        current_log_dir = os.path.join("log", f"{_get_main_file_name()}_{timestamp('%Y-%m-%d')}")
    # remove dots from screenshot format in case user provided ".png" instead of "png"
    screenshot_format = screenshot_format.replace(".", "")
    error_imgs = sorted(glob.glob(f"{current_log_dir}\\*.{screenshot_format}"), key=os.path.getmtime)
    error_img_path_log = error_imgs[-1 - offset]
    if not output_file_name:
        # if no output file name was provided, use name of the original image
        output_file_name = os.path.basename(error_img_path_log)
    else:
        output_file_name += f".{screenshot_format}"
    error_img_path_output = os.path.join(output_dir, output_file_name)
    shutil.copyfile(error_img_path_log, error_img_path_output)
    return error_img_path_output



"""Information about urpatimeout module."""
__title__ = "urpautils"
__description__ = "Module containig universal functions used with UltimateRPA robots"
__version__ = "0.1.3"
__author__ = "anerold"
__author_email__ = "support@ultimaterpa.cz"
__url__ = "https://github.com/ultimaterpa/urpautils"
__license__ = "MIT"



"""Module containing universal functions for csv files manipulation that can be used with urpa robots"""

import csv
import logging
import os
import sys

from itertools import islice
from typing import Optional, Iterable, Generator

logger = logging.getLogger(__name__)


def csv_append_row(
    file_path: str, row: Iterable, newline: str = "", encoding: str = "utf-8", delimiter: str = ";"
) -> None:
    """Appends a row to an existing csv file.

    :param file_path:    path
    :param row:          iterable of columns in one row
    :param newline:      newline character
    :param encoding:     encoding of the csv file
    :param delimiter:    csv delimiter
    :return:             None
    """
    if not os.path.isfile(file_path):
        raise FileNotFoundError(f"File '{file_path}' does not exist")
    with open(file_path, "a", newline=newline, encoding=encoding) as csv_file:
        csv_writer = csv.writer(csv_file, delimiter=delimiter)
        csv_writer.writerow(row)


def csv_create_file(
    file_path: str,
    header: Optional[Iterable] = None,
    newline: str = "",
    encoding: str = "utf-8",
    delimiter: str = ";",
) -> None:
    """Creates new csv file and writes its header if provided

    :param file_path:     path
    :param header:        optional iterable of columns to write as a header
    :param newline:       newline character
    :param encoding:      encoding of the csv file
    :param delimiter:     csv delimiter
    return None
    """
    if os.path.isfile(file_path):
        raise FileExistsError(f"File '{file_path}' already exists")
    if header:
        # raises TypeError if 'header' is not iterable
        iter(header)

    with open(file_path, "w", newline=newline, encoding=encoding) as csv_file:
        if header:
            csv_writer = csv.writer(csv_file, delimiter=delimiter)
            csv_writer.writerow(header)


def csv_read_rows(
    file_path: str,
    start_row_index: int = 0,
    end_row_index: int = sys.maxsize,
    newline: str = "",
    encoding: str = "utf-8",
    delimiter: str = ";",
) -> Generator[list, None, None]:
    """Generates rows of a csv file.
    If start_row_index is provided it starts generating from that row index
    If end_row_index is provided it stops generating at that index
    If no start_row_index and end_row_index is provided it yields all rows

    :param file_path:     path
    :param row_index:     index of the row to return
    :return:              list of columns in the row
    """
    if not os.path.isfile(file_path):
        raise FileNotFoundError(f"File '{file_path}' does not exist")
    with open(file_path, newline=newline, encoding=encoding) as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=delimiter)
        for row in islice(csv_reader, start_row_index, end_row_index):
            yield row



"""Module containing functions dependent on the urpa library that can be used with urpa robots"""

import logging

from typing import Tuple, Union, Sequence, Optional

import urpa
import urpatimeout  # TODO mention this in setup.py


cf = urpa.condition_factory()
logger = logging.getLogger(__name__)


def parallel_search(
    app: urpa.App, *conditions: Union[urpa.Condition, Sequence[urpa.Condition]], wait: int = 10000
) -> Tuple[int, list]:
    """Searches for multiple elements. Return first one found

    :param app:            urpa app
    :param *conditions:    cf conditions of the elements
    :param wait:           timeout in ms
    :return:               tuple (index of the element found, element)
    """
    t = urpatimeout.Timeout(wait)
    while not t.is_expired():
        for index, condition in enumerate(conditions):
            elements = app.find_all(condition, 0)
            if elements:
                return index, elements
    raise urpa.ElementNotFoundError("Parallel_search: No elemets found for the conditions!")


def check_elements(
    app: urpa.App,
    *conditions: Union[urpa.Condition, Sequence[urpa.Condition]],
    timeout: Optional[int] = None,
) -> None:
    """Searches for control elements

    :param app:            urpa app
    :param *conditions:    cf conditions of the element(s)
    :param timeout:        optional timeout in ms (urpa.default_timeout is used if None)
    :return:               None
    """
    logger.debug("Searching for control elements")
    for condition in conditions:
        if timeout is None:
            app.find_first(condition)
        else:
            app.find_first(condition, timeout=timeout)
    logger.debug("Control elements found")


def robot_setup(
    screen_resolution: Tuple[int, int, int],
    default_timeout: Optional[int] = None,
    default_screenshot_format: str = "png",
    debug_mode: bool = False,
) -> None:
    """Initiates urpa robot

    :param screen_resolution:          tuple (width, heigh, bit depth)
    :param default_timeout:            optional default timeout for element searching. urpa.default_timeout is used if None
    :param default_screenshot_format:  png or bmp
    :param debug_mode:                 bool robot debug mode
    :return:                           None
    """
    possible_screenshot_formats = ("png", "bmp")
    if not default_screenshot_format in possible_screenshot_formats:
        raise ValueError(
            f"Invalid screenhsot format '{default_screenshot_format}'. Please use one of the following: '{possible_screenshot_formats}'"
        )
    urpa.bring_to_foreground()
    if default_timeout is not None:
        urpa.set_default_timeout(default_timeout)
    try:
        urpa.check_screen_resolution(screen_resolution[0], screen_resolution[1], screen_resolution[2])
    except ValueError:
        urpa.set_screen_resolution(screen_resolution[0], screen_resolution[1], screen_resolution[2])
    urpa.default.screenshot_format = default_screenshot_format
    urpa.set_debug_mode(debug_mode)



import pathlib
from setuptools import setup

here = pathlib.Path(__file__).parent
readme = (here / "README.md").read_text(encoding="utf-8")

about = {}
with open(here / "urpautils" / "__about__.py") as f:
    exec(f.read(), about)

setup(
    name=about["__title__"],
    version=about["__version__"],
    description=about["__description__"],
    long_description=readme,
    long_description_content_type="text/markdown",
    url=about["__url__"],
    author=about["__author__"],
    author_email=about["__author_email__"],
    license=about["__license__"],
    keywords="Robotic Process Automation, RPA, UltimateRPA, utilities",
    packages_data={"urpautils": ["py.typed"]},
    packages=["urpautils"],
    python_requires=">=3.7",
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.7",
        "Operating System :: Microsoft :: Windows",
        "License :: OSI Approved :: MIT License",
    ],
    zip_safe=False,
    include_package_data=True,
)



"""Module containing universal functions that can be used with urpa robots"""

import datetime
import logging
import re
import smtplib
import subprocess
import time
from typing import List, Optional

from email import charset
from email.header import Header
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText


logger = logging.getLogger(__name__)


def timestamp(format: str = "%d.%m.%Y %H:%M:%S", pad_date_with_zeros: bool = True) -> str:
    """Returns formatted timestamp

    :param format:               timestamp format
    :param pad_date_with_zeros:  pad one-digit numbers of day and month with leading zero
    :return:                     timestamp string
    """
    if not pad_date_with_zeros:
        format = format.replace("%m", "%#m").replace("%d", "%#d")
    return datetime.datetime.now().strftime(format)


def add_long_path_prefix(path: str) -> str:
    r"""Returns original path with prefix to avoid win32api 260 char limit
    C:\Temp -> \\?\C:\Temp
    \\net\folder -> \\?\UNC\net\folder
    https://docs.microsoft.com/en-us/windows/win32/fileio/maximum-file-path-limitation?tabs=cmd

    :param path:          path
    :return:              original path with prefix
    """
    local_prefix = "\\\\?\\"
    network_prefix = "\\\\?\\UNC\\"
    if path[0:2] == "\\\\":
        return f"{network_prefix}{path[2:]}"
    return f"{local_prefix}{path}"


def kill_app(image_name: str, check: bool = False) -> None:
    """Kills app and its tree

    :param image_name:      app image name
    :param check:           check status code of the subprocess.run. Raises Exception if non-zero
    :return:                None
    """
    subprocess.run(("taskkill", "/F", "/IM", f"{image_name}"), check=check)


def send_email_notification(
    email_sender: str,
    recipients: List[str],
    recipients_copy: List[str],
    subject: str,
    body: str,
    smtp_server: str,
    smtp_port: int = 0,
) -> None:
    """Sends an e-mail

    :param email_sender:    sender of this e-mail
    :param recipients:      list of addresses of recipients
    :param recipients_copy: list of addresses of recipients of a copy
    :param subject:         subject of the e-mail
    :param body:            body of the e-mail
    :param smtp_server:     smtp server
    :param smtp_port:       optional port for the smtp server. smtplib.SMTP_PORT (=25) is used if not provided
    :return:                None
    """
    charset.add_charset("utf-8", charset.QP, charset.QP, "utf-8")
    mail = MIMEMultipart("alternative")
    mail["Subject"] = Header(subject, "utf-8")
    mail["From"] = email_sender
    mail["To"] = ", ".join(recipients)
    mail["Cc"] = ", ".join(recipients_copy)
    text = body
    html = "<html><body>" + body + "</body></html>"
    txt_message = MIMEText(text.encode("utf-8"), "plain", "utf-8")  # type: ignore
    html_message = MIMEText(html.encode("utf-8"), "html", "utf-8")  # type: ignore
    mail.attach(txt_message)
    mail.attach(html_message)
    logger.info(f"Sending e-mail to '{recipients}', copy: '{recipients_copy}'")
    logger.debug(f"Subject: '{subject}', body: '{body}'")
    sender = smtplib.SMTP(smtp_server, smtp_port)
    sender.set_debuglevel(1)
    sender.sendmail(email_sender, recipients + recipients_copy, mail.as_string())
    sender.quit()
    logger.info("E-mail sent")


def get_birth_date(number: str) -> datetime.date:
    """Returns birth date calculated from personal identification number

    :param number:   string in format xxxxxxxxxx
    :return:         datetime.date object (or raises ValueError)
    """
    if not len(number) in (9, 10):
        raise ValueError("Invalid length")
    year = 1900 + int(number[0:2])
    # females have 50 added to the month value, 20 is added when the serial
    # overflows (since 2004)
    month = int(number[2:4]) % 50 % 20
    day = int(number[4:6])
    # 9 digit numbers were used until January 1st 1954
    if len(number) == 9:
        if year >= 1980:
            year -= 100
        if year > 1953:
            raise ValueError("No 9 digit birth numbers after 1953.")
    elif year < 1954:
        year += 100
    # this can also raise ValueError
    return datetime.date(year, month, day)


def verify_rc(number: str) -> bool:
    """Check whether 'Rodné číslo' is valid
    https://phpfashion.com/jak-overit-platne-ic-a-rodne-cislo

    :param number:   string in format xxxxxxxxxx or xxxxxx/xxxx
    :return:         bool
    """
    # check whether 'number' contains '/' and if its at 5th last place of the input (i.e. xxxxxx/xxxx)
    # if yes, remove it
    if number.count("/"):
        if number.count("/") == 1 and len(number) - number.index("/") == 5:
            number = number.replace("/", "")
        else:
            # the number has '/' but not in the right place or there are more than one
            return False

    if len(number) not in (9, 10):
        return False
    try:
        birth_date = get_birth_date(number)
    except ValueError as err:
        logger.debug(f"RC '{number}' not valid, could not verify birth date: '{str(err)}'")
        return False
    # check the check digit (10 digit numbers only)
    if len(number) == 10:
        check = int(number[:-1]) % 11
        # before 1985 the checksum could be 0 or 10
        if birth_date < datetime.date(1985, 1, 1):
            check = check % 10
        if number[-1] != str(check):
            return False
    return True


def verify_ico(bin: str) -> bool:
    """Check whether IČO number is valid
    https://phpfashion.com/jak-overit-platne-ic-a-rodne-cislo

    :param bin:   string of the BIN number
    :retun:       bool
    """
    if isinstance(bin, str):
        if not bin.isnumeric() or len(bin) > 8:
            return False
        # pad with zeros from left to length 8 (123456 -> 00123456)
        bin = bin.rjust(8, "0")
    else:
        raise ValueError(f"'BIN' must be an instance of 'str', not '{type(bin)}'")

    a = 0
    for i in range(7):
        a += int(bin[i]) * (8 - i)

    a %= 11
    if a == 0:
        c = 1
    elif a == 1:
        c = 0
    else:
        c = 11 - a

    return int(bin[7]) == c


def clear_ie_cache(check: bool = False) -> None:
    r"""Clears Internet Explorer cache and makes registry change so it does not open 'customize' window on first run
    RunDll32.exe InetCpl.cpl,ClearMyTracksByProcess 255
    reg ADD "HKEY_CURRENT_USER\Software\Policies\Microsoft\Internet Explorer\Main" /v DisableFirstRunCustomize /d 1 /t REG_DWORD /f
    reg ADD "HKEY_CURRENT_USER\Software\Policies\Microsoft\Internet Explorer\Main" /v RunOnceComplete /d 1 /t REG_DWORD /f
    reg ADD "HKEY_CURRENT_USER\Software\Policies\Microsoft\Internet Explorer\Main" /v RunOnceHasShown /d 1 /t REG_DWORD /f

    :param check:    check status code of the subprocess.run. Raises Exception if non-zero
    """
    subprocess.run(("RunDll32.exe", "InetCpl.cpl,ClearMyTracksByProcess", "255"), check=check)
    time.sleep(5)
    subprocess.run(
        (
            "reg",
            "ADD",
            r"HKEY_CURRENT_USER\Software\Policies\Microsoft\Internet Explorer\Main",
            "/v",
            "DisableFirstRunCustomize",
            "/d",
            "1",
            "/t",
            "REG_DWORD",
            "/f",
        ),
        check=check,
    )
    subprocess.run(
        (
            "reg",
            "ADD",
            r"HKEY_CURRENT_USER\Software\Policies\Microsoft\Internet Explorer\Main",
            "/v",
            "RunOnceComplete",
            "/d",
            "1",
            "/t",
            "REG_DWORD",
            "/f",
        ),
        check=check,
    )
    subprocess.run(
        (
            "reg",
            "ADD",
            r"HKEY_CURRENT_USER\Software\Policies\Microsoft\Internet Explorer\Main",
            "/v",
            "RunOnceHasShown",
            "/d",
            "1",
            "/t",
            "REG_DWORD",
            "/f",
        ),
        check=check,
    )


def get_app_pid(
    app_name: str,
    pids_to_exclude: List[Optional[int]] = [],
    number_of_retries: int = 3,
    wait_before_next_try: int = 10000,
) -> int:
    """Finds all PIDs of an 'app_name' application and returns first one

    :param app_name:              image name of the app
    :param pids_to_exclude:       list of PIDs that are filtered out
    :param number_of_retries:     number of retries
    :param wait_before_next_try:  amount of miliseconds to wait before next try
    :return:                      int of the first PID found
    """
    if not isinstance(pids_to_exclude, list):
        raise ValueError(f"'pids_to_exclude' must be 'list', not '{type(pids_to_exclude)}'")
    for retry in range(1, number_of_retries + 1):
        logger.info(f"Try '{retry}' of finding PID for application '{app_name}'")
        tasklist_command = f'tasklist /FI "IMAGENAME eq {app_name}*"'
        for pid_to_exclude in pids_to_exclude:
            tasklist_command += f' /FI "PID ne {pid_to_exclude}"'
        tasklist_output = subprocess.check_output(tasklist_command)
        # find first number surrounded by whitespaces
        try:
            pid = int(re.search(rb"\s+(\d+)\s+", tasklist_output).group(1))  # type: ignore
        except AttributeError:
            logger.warning("PID not found")
            time.sleep(wait_before_next_try / 1000)
            continue
        logger.info(f"PID found: {pid}")
        return pid
    raise RuntimeError(f"Unable to find PID for image name '{app_name}'")
